"""
This file is part of nucypher.

nucypher is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

nucypher is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with nucypher.  If not, see <https://www.gnu.org/licenses/>.
"""


from __future__ import absolute_import, division, print_function

__all__ = [
    "__title__", "__summary__", "__version__", "__author__", "__email__", "__license__", "__copyright__", "__url__"
]

__title__ = "nucypher"

__url__ = "https://github.com/nucypher/nucypher"

__summary__ = 'A proxy re-encryption network to empower privacy in decentralized systems.'

__version__ = "0.1.0-alpha.27"

__author__ = "NuCypher"

__email__ = "dev@nucypher.com"

__license__ = "GNU Affero General Public License, Version 3"

__copyright__ = 'Copyright (C) 2019 NuCypher'
