

#
# Contract Names
#

DISPATCHER_CONTRACT_NAME = 'Dispatcher'
LIBRARY_LINKER_CONTRACT_NAME = "UserEscrowLibraryLinker"
NUCYPHER_TOKEN_CONTRACT_NAME = 'NuCypherToken'
STAKING_ESCROW_CONTRACT_NAME = 'StakingEscrow'
POLICY_MANAGER_CONTRACT_NAME = 'PolicyManager'
USER_ESCROW_PROXY_CONTRACT_NAME = 'UserEscrowProxy'
USER_ESCROW_CONTRACT_NAME = 'UserEscrow'
ADJUDICATOR_CONTRACT_NAME = 'Adjudicator'
WORKLOCK_CONTRACT_NAME = 'WorkLock'
